from enum import IntEnum

from energyweb.interfaces import ExternalData, IntegrationPoint


class EnergyUnit(IntEnum):
    """
    Possible units for effort and billing purposes. Determines the convertion algorithms.
    """
    JOULES = 0
    WATT_HOUR = 1
    KILOWATT_HOUR = 2
    MEGAWATT_HOUR = 3
    GIGAWATT_HOUR = 4


class EnergyData(ExternalData):
    """
    Standard for collected energy and power data, to be transformed into mintable data.
    """

    def __init__(self, device, access_epoch, raw, energy, measurement_epoch):
        """
        Minimum set of data for power measurement logging.
        :param device: Metadata about the measurement device. EnergyDevice
        :param access_epoch: Time the external API was accessed
        :param raw: Raw data collected
        :param energy: Measured energy at the source converted to watt-hours
        :param measurement_epoch: Time of measurement at the source
        """
        self.device = device
        self.measurement_epoch = measurement_epoch
        self.energy = energy
        ExternalData.__init__(self, access_epoch, raw)


class EnergyDevice(IntegrationPoint):
    """
    Energy device or api abstraction. Can be smart-meters, battery controllers, inverters, gateways, clouds, so on.
    """

    def __init__(self, manufacturer, model, serial_number, energy_unit, is_accumulated, latitude=None, longitude=None):
        """
        :param manufacturer: EnergyAsset Manufacturer
        :param model: EnergyAsset model
        :param serial_number: EnergyAsset Serial Number
        :param latitude: EnergyAsset geolocation latitude
        :param longitude: EnergyAsset geolocation longitude
        :param energy_unit: Energy unity
        :param is_accumulated: Flags if the api provides accumulated power or hourly production
        """
        self.manufacturer = manufacturer
        self.model = model
        self.serial_number = serial_number
        self.latitude = latitude
        self.longitude = longitude
        self.energy_unit = EnergyUnit[energy_unit.upper()]
        self.is_accumulated = is_accumulated

    def read_state(self, *args, **kwargs) -> EnergyData:
        """
        Establishes a connection to the integration medium and returns the latest state
        :rtype: EnergyData
        """
        raise NotImplementedError

    def write_state(self, *args, **kwargs) -> EnergyData:
        """
        Establishes a connection to the integration medium and returns the latest state
        :rtype: EnergyData
        """
        raise NotImplementedError

    @staticmethod
    def to_mwh(energy: int, unit):
        """
        Converts energy measured value in predefined unit to Megawatt-Hour
        :param unit: EnergyUnit
        :param energy: Value measured at the source
        :return: Value converted to MWh
        :rtype: float
        """
        convert = lambda x: float(energy) / x
        return {
            EnergyUnit.WATT_HOUR: convert(10 ** 6),
            EnergyUnit.KILOWATT_HOUR: convert(10 ** 3),
            EnergyUnit.MEGAWATT_HOUR: convert(1),
            EnergyUnit.GIGAWATT_HOUR: convert(1 * 10 ** -3),
            EnergyUnit.JOULES: convert(1 * 2.77778 * 10 ** -10),
        }.get(unit)

    @staticmethod
    def to_wh(energy: int, unit: EnergyUnit):
        """
        Converts energy measured value in predefined unit to Watt-Hour
        :param unit: EnergyUnit
        :param energy: Value measured at the source
        :return: Value converted to MWh
        :rtype: int
        """
        convert = lambda x: int(float(energy * x))
        return {
            EnergyUnit.WATT_HOUR: convert(1),
            EnergyUnit.KILOWATT_HOUR: convert(10 ** 3),
            EnergyUnit.MEGAWATT_HOUR: convert(10 ** 6),
            EnergyUnit.GIGAWATT_HOUR: convert(10 ** 9),
            EnergyUnit.JOULES: convert(1 * 2.77778 * 10 ** -1),
        }.get(unit)



"""
 _
| |                  |
| |   __   _  _    __|
|/ \_/  \_/ |/ |  /  |
 \_/ \__/   |  |_/\_/|_/

Bond - Your favorite library for logging energy data on the blockchain
"""

import energyweb.config as config

from energyweb.interfaces import Serializable, ExternalData, IntegrationPoint, BlockchainClient
from energyweb.log import Logger
from energyweb.dispatcher import App, Task
from energyweb.carbonemission import CarbonEmissionData
from energyweb.eds.interfaces import EnergyUnit, EnergyData, EnergyDevice
from energyweb.smart_contract.origin_v1 import ConsumedEnergy, ProducedEnergy, OriginV1
from energyweb.smart_contract.interfaces import EVMSmartContractClient
from energyweb.storage import DiskStorage


__name__ = 'energyweb'
__author__ = 'Paul Depraz <github.com/cerealkill>'
__repository__ = 'github.com/energywebfoundation/ew-link-bond'
__status__ = "pre-alpha"
__version__ = "0.4.0"
__date__ = "14 December 2018"



import datetime
import inspect


class Serializable(object):
    """
    Object serialization helper
    """
    def __iter__(self):
        for attr, value in self.__dict__.items():
            if isinstance(value, datetime.datetime):
                iso = value.isoformat()
                yield attr, iso
            elif hasattr(value, '__iter__'):
                if hasattr(value, 'pop'):
                    a = []
                    for item in value:
                        if hasattr(item, '__iter__'):
                            a.append(dict(item))
                        else:
                            a.append(item)
                    yield attr, a
                else:
                    yield attr, dict(value)
            else:
                yield attr, value

    def to_dict(self):
        result = {}
        init = getattr(self, '__init__')
        for parameter in inspect.signature(init).parameters:
            att = getattr(self, parameter)
            if isinstance(att, list) or isinstance(att, set):
                att = [self.to_dict_or_self(o) for o in att]
            if isinstance(att, (datetime.datetime, datetime.date)):
                att = att.isoformat()
            result[parameter] = self.to_dict_or_self(att)
        return result

    @staticmethod
    def to_dict_or_self(obj):
        to_dict = getattr(obj, 'to_dict', None)
        if to_dict:
            return to_dict()
        else:
            return obj


class ExternalData(Serializable):
    """
    Encapsulates collected data in a traceable fashion
    """

    def __init__(self, access_epoch, raw):
        """
        :param access_epoch: Time the external API was accessed
        :param raw: Raw data collected in the device
        """
        self.access_epoch = access_epoch
        self.raw = raw


class IntegrationPoint:
    """
    Interface to enforce correct return type and standardized naming
    """
    def read_state(self, *args, **kwargs) -> ExternalData:
        """
        Establishes a connection to the integration medium and returns the latest state
        :rtype: ExternalData
        """
        raise NotImplementedError

    def write_state(self, *args, **kwargs) -> ExternalData:
        """
        Establishes a connection to the integration medium and persists data
        :rtype: ExternalData
        """
        raise NotImplementedError


class BlockchainClient:
    """
    Abstract blockchain client abstraction
    """

    def is_synced(self) -> bool:
        """
        Compares latest block from peers with client's last synced block.
        :return: Synced status
        :rtype: bool
        """
        raise NotImplementedError

    def call(self, address: str, contract_name: str, method_name: str, password: str, args=None) -> dict:
        """
        Calls a method in a smart-contract
        Sends a transaction to the Blockchain and awaits for mining until a receipt is returned.
        :param address: Contract address
        :param contract_name: Name of the contract in contracts
        :param method_name: Use the same name as found in the contract abi
        :param password: String of the raw password
        :param args: Method parameters
        :return: Transaction receipt
        :rtype: dict
        """
        raise NotImplementedError

    def send(self, address: str, contract_name: str, method_name: str, password: str, args=None) -> dict:
        """
        Send a transaction to execute a method in a smart-contract
        Sends a transaction to the Blockchain and awaits for mining until a receipt is returned.
        :param address: Contract address
        :param contract_name: Name of the contract in contracts
        :param method_name: Use the same name as found in the contract abi
        :param password: String of the raw password
        :param args: Method parameters
        :return: Transaction receipt
        :rtype: dict
        """
        raise NotImplementedError

    def mint(self, energy: ExternalData) -> dict:
        """
        Mint the measured energy in the blockchain smart-contract
        """
        raise NotImplementedError




"""
Configuration file parser and app descriptor
"""
import importlib
from collections import namedtuple
from enum import Enum

from energyweb.carbonemission import CarbonEmissionAPI
from energyweb.eds.interfaces import EnergyDevice
from energyweb.smart_contract.interfaces import EVMSmartContractClient

Module = namedtuple('Module', ['module', 'class_name', 'parameters'])


class MODULES(Enum):
    OriginAPIv1 = Module('energyweb.smart_meter.energy_api', 'OriginAPIv1', {})
    OriginAPIv1Test = Module('energyweb.smart_meter.energy_api', 'OriginAPIv1', {'url': '', 'source': 'consumed', 'device_id': 0})
    ConsumerTask = Module('energyweb.smart_meter.energy_api', 'OriginAPIv1', {})
    ProducerTask = Module('energyweb.smart_meter.energy_api', 'OriginAPIv1', {})


class EnergyAssetConfiguration:
    """
    Represent Energy Asset configurations for producer and consumers
    """
    def __init__(self, name: str, energy_meter: EnergyDevice, asset: EVMSmartContractClient):
        if not isinstance(energy_meter, EnergyDevice):
            raise ConfigurationFileError('Energy_meter must be of an EnergyDevice class or subclass.')
        if len(name) < 2:
            raise ConfigurationFileError('Name must be longer than two characters.')
        self.name = name
        self.meter = energy_meter
        self.asset = asset


class GreenEnergyProducerConfiguration(EnergyAssetConfiguration):
    """
    Represents special case of energy producer
    """
    def __init__(self, name: str, energy_meter: EnergyDevice, carbon_emission: CarbonEmissionAPI,
                 asset: EVMSmartContractClient):
        if not isinstance(carbon_emission, CarbonEmissionAPI):
            raise ConfigurationFileError('Carbon_emission must be of an CarbonEmissionAPI class or subclass.')
        self.carbon_emission = carbon_emission
        super().__init__(name=name, energy_meter=energy_meter, asset=asset)


class BondConfiguration:
    """
    Bond configuration file format
    :version v0.3.7
        energyweb: 1.0
        info:
          title: App Title
          command: My very nice energyweb embedded app
          version: 1.0 alpha
        task_config:
          debugger: on
          verbosity: low
        tasks:
    """
    class AppInfo:
        def __init__(self, title: str, description: str, version: str):
            self.title = title
            self.description = description
            self.version = version

    class AppConfig:
        def __init__(self, debugger: str, verbosity: str, **kwargs):
            self.debugger = debugger
            self.verbosity = verbosity
            self.__dict__.update(kwargs)

    class Submodule:
        def __init__(self, module: str, **kwargs):
            self.module = module
            self.__dict__.update(kwargs)

    def __init__(self, energyweb: str, info: dict, config: dict, tasks: [dict]):
        if len(tasks) < 1:
            raise ConfigurationFileError('Should contain at least one valid task.')
        self.energyweb = energyweb
        self.tasks = tasks
        self.config = self.AppConfig(config)
        self.info = self.AppInfo(info)

# TODO: Refactor origin config and decide if it lives here or in origin app
# def parse_bond_app(configuration_file_path: str) -> BondConfiguration:
#     """
#     :param configuration_file_path: Configuration file parsed as a dictionary.
#     :return: BondConfiguration
#     """
#     from ruamel.yaml import YAML
#     yaml = YAML()  # default, if not specfied, is 'rt' (round-trip)
#     try:
#         config = yaml.load(open(configuration_file_path))
#
#         version = config['bond_version']
#         tasks = [__parse_instance(task) for task in config['tasks']]
#         configuration = BondConfiguration(bond_version=version, tasks=tasks)
#     except Exception:
#         raise ConfigurationFileError
#     return configuration


class ConfigurationFileError(Exception):
    """
    Custom Exception to deal with non-conforming configuration files
    """
    def __init__(self, msg: str = None):
        if not msg:
            super().__init__('Configuration file contain errors. Please provide valid Consumer and/or Producer data.')
        super().__init__(msg)


class EnergywebAppConfiguration:
    """
    Abstract class
    """
    pass


class CooV1ConsumerConfiguration(EnergywebAppConfiguration):
    """
    Represent Energy Asset configurations for producer and consumers
    """
    def __init__(self, name: str, energy_meter: EnergyDevice, smart_contract: EVMSmartContractClient):
        if not isinstance(energy_meter, EnergyDevice):
            raise ConfigurationFileError('Energy_meter must be of an EnergyDevice class or subclass.')
        if len(name) < 2:
            raise ConfigurationFileError('Name must be longer than two characters.')
        self.name = name
        self.energy_meter = energy_meter
        self.smart_contract = smart_contract


class CooV1ProducerConfiguration(CooV1ConsumerConfiguration):
    """
    Represents special case of energy producer
    """
    def __init__(self, name: str, energy_meter: EnergyDevice, carbon_emission: CarbonEmissionAPI,
                 smart_contract: EVMSmartContractClient):
        if not isinstance(carbon_emission, CarbonEmissionAPI):
            raise ConfigurationFileError('Carbon_emission must be of an CarbonEmissionAPI class or subclass.')
        self.carbon_emission = carbon_emission
        super().__init__(name=name, energy_meter=energy_meter, smart_contract=smart_contract)


class CooV1Configuration(EnergywebAppConfiguration):
    """
    When using one file configuration for multiple assets
    """
    def __init__(self, consumption: [CooV1ConsumerConfiguration], production: [CooV1ConsumerConfiguration]):
        self.consumption = consumption
        self.production = production


def parse_coo_v1(raw_configuration_file: dict) -> CooV1Configuration:
    """
    Read and parse Certificatate of Orign v1 (release A) configuration dictionary into structured class instances.
    :param raw_configuration_file: Configuration file parsed as a dictionary.
    :return: Configuration instance
    """
    if not isinstance(raw_configuration_file, dict):
        raise AssertionError("Configuration json must be deserialized first.")
    is_consuming = 'consumers' in raw_configuration_file
    is_producing = 'producers' in raw_configuration_file
    if not is_consuming and not is_producing:
        raise ConfigurationFileError

    consumption = [__parse_item(config) for config in raw_configuration_file['consumers']]
    production = [__parse_item(config) for config in raw_configuration_file['producers']]
    return CooV1Configuration(consumption, production)


def parse_single_asset(raw_configuration_file: dict) -> CooV1ConsumerConfiguration:
    """
    Read and parse single producer or consumer configuration dictionary into structured class instances.
    :param raw_configuration_file: Configuration file parsed as a dictionary.
    :return: EnergyConsumerConfiguration or EnergyProducerConfiguration
    """
    if not isinstance(raw_configuration_file, dict):
        raise AssertionError("Configuration json must be deserialized first.")
    configuration = __parse_item(raw_configuration_file)


def __parse_item(config_item: dict):
    """
    Parse each item of the consumers and producers list.
    :param config_item: Item of the list
    :return: Either Producer or Consumer configuration depending if carbon-emission key is present
    """
    item = {
        'energy_meter': __parse_instance(config_item['energy-meter']),
        'smart_contract': __parse_instance(config_item['smart-contract']),
        'name': config_item['name']
    }
    if 'carbon-emission' not in config_item:
        return CooV1ConsumerConfiguration(**item)
    else:
        item['carbon_emission'] = __parse_instance(config_item['carbon-emission'])
        return CooV1ProducerConfiguration(**item)


def __parse_instance(submodule: dict) -> object:
    """
    Reflection algorithm to dynamically load python modules referenced on the configuration json.
    :param submodule: Configuration dict must have the keys 'module', 'class_name', 'class_parameters'.
    :return: Class instance as in task_config file.
    """
    module_instance = importlib.import_module(submodule['module'])
    class_obj = getattr(module_instance, submodule['class_name'])
    class_instance = class_obj(**submodule['class_parameters'])
    return class_instance


"""
__Storage__ supports EWF's Origin release A log storage, designed to record a sequence of _off-chain_ files by updating the previous file contents SHA hash with the next. It is particularly useful to enforce data integrity, by comparing the sequence of raw smart-energy_meter readings with the sequence registered _on-chain_.
"""
import os
import json
import pickle
import hashlib
import datetime
import base58

from energyweb.interfaces import Serializable


class ChainFile:
    """
    List element
    """
    def __init__(self, file: str, timestamp: datetime.datetime):
        self.file = file
        self.timestamp = timestamp


class ChainLink:
    """
    List link
    """
    def __init__(self, data: ChainFile, last_link: object):
        self.data = data
        self.last_link = last_link

    def __next__(self):
        return self.last_link


class DiskStorage:
    """
    Saves a pickle with the data in chain format.
    """
    def __init__(self, chain_file_name: str, path_to_files: str):
        """
        :param chain_file_name:
        :param path_to_files:
        """
        self.chain_file = os.path.join(path_to_files, chain_file_name)
        self.path = path_to_files
        os.makedirs(path_to_files, exist_ok=True)
        if not os.path.exists(self.chain_file):
            self.__memory = None
            return
        try:
            self.__memory = pickle.load(open(self.chain_file, 'rb'))
        except EOFError:
            self.__memory = None

    @property
    def chain(self) -> ChainLink:
        return self.__memory

    @chain.setter
    def chain(self, chain_link: ChainLink):
        if chain_link is not None:
            raise AttributeError
        self._chain_append(chain_link)

    def add_to_chain(self, data: Serializable) -> str:
        """
        Add new file to chain.
        :param data: Data to store
        :return: File name string
        """
        data_file_name = self._save_file(data)
        chain_data = ChainFile(data_file_name, datetime.datetime.now())
        new_link = ChainLink(data=chain_data, last_link=self.chain)
        self._chain_append(new_link)
        self._save_memory()
        return data_file_name

    def get_last_hash(self) -> str:
        """
        Get hash of the last chain file.
        :return: Base58 hash string
        """
        if self.chain:
            # sha3 = hashlib.sha3_256()
            sha3 = hashlib.sha1()
            sha3.update(open(self.chain.data.file, 'rb').read())
            base58_digest = base58.b58encode(sha3.digest())
            return 'Qm' + base58_digest.decode()
        else:
            return '0x0'

    def _chain_append(self, chain_link: ChainLink):
        self.__memory = chain_link
        self._save_memory()

    def _save_memory(self):
        pickle.dump(self.__memory, open(self.chain_file, 'wb'), protocol=pickle.HIGHEST_PROTOCOL)

    def _save_file(self, data):
        if not os.path.exists(self.path):
            os.makedirs(self.path)
        file_name_mask = os.path.join(self.path, '%Y-%m-%d-%H:%M:%S.json')
        file_name = datetime.datetime.now().strftime(file_name_mask)
        with open(file_name, 'w+') as file:
            json.dump(data.to_dict(), file)
        return file_name



"""
Library containing the implementations of smart-energy_meter simulator integration classes
"""
import time
import random

from energyweb.eds.interfaces import EnergyDevice, EnergyData

class EnergyMeterSimulator(EnergyDevice):
    """
    Data logger simulator. It will return a pseudo-random incremental number in every iteration
    """
    def __init__(self):
        self.memory = random.randint(1, 20)

    def read_state(self) -> EnergyData:
        access_epoch = int(time.time())
        device = EnergyDevice(
            manufacturer='Slock.it',
            model='Virtual Energy Meter',
            serial_number='0001000',
            latitude='1.123',
            longitude='1.321')
        kwh_power = random.randint(self.memory, (self.memory + 1) + 20)
        measurement_epoch = int(time.time())
        device_str = device.manufacturer + device.model + device.serial_number
        raw = str(device_str + str(access_epoch) + str(kwh_power) + str(measurement_epoch))
        return EnergyData(device=device, access_epoch=access_epoch, raw=raw, energy=kwh_power,
                          measurement_epoch=measurement_epoch)

    def write_state(self, *args, **kwargs):
        raise NotImplementedError



"""
input.eumel

Library containing the implementations of Eumel DataLogger integration classes
"""
import time
import requests

from xml.etree import ElementTree
from energyweb.eds import EnergyUnit, EnergyData
from energyweb.eds.interfaces import EnergyDevice


class DataLoggerV1(EnergyDevice):
    """
    Eumel DataLogger api v1.0 access implementation
    """

    def __init__(self, ip, user, password):
        """
        :param ip: Data loggers network IP
        :param user: User configured on the devices
        :param password: Password for this user
        """
        self.eumel_api_url = ip + '/rest'
        self.auth = (user, password)
        super().__init__(manufacturer='Verbund', model='Eumel v1', serial_number=None, energy_unit=EnergyUnit.WATT_HOUR,
                         is_accumulated=True)

    def read_state(self, path=None) -> EnergyData:
        if path:
            tree = ElementTree.parse(path)
            with open(path) as file:
                raw = file.read()
        else:
            http_packet = requests.get(self.eumel_api_url, auth=self.auth)
            raw = http_packet.content.decode()
            tree = ElementTree.parse(raw)
        tree_root = tree.getroot()
        tree_header = tree_root[0].attrib
        tree_leaves = {child.attrib['id']: child.text for child in tree_root[0][0]}
        device = EnergyDevice(
            manufacturer=tree_header['man'],
            model=tree_header['mod'],
            serial_number=tree_header['sn'],
            energy_unit=self.energy_unit,
            is_accumulated=self.is_accumulated)
        access_epoch = int(time.time())
        time_format = '%Y-%m-%dT%H:%M:%SZ'
        energy = float(tree_leaves['TotWhImp'].replace('.', ''))
        mwh_energy = self.to_mwh(energy, self.energy_unit)
        measurement_epoch = int(time.mktime(time.strptime(tree_header['t'], time_format)))
        return EnergyData(asset=device, access_epoch=access_epoch, raw=raw, measurement_epoch=measurement_epoch,
                          energy=mwh_energy)

    def write_state(self, *args, **kwargs):
        raise NotImplementedError


class DataLoggerV2d1d1(EnergyDevice):
    """
    Eumel DataLogger api v2.1.1 access implementation
    """

    def __init__(self, ip, user, password):
        """
        :param ip: Data loggers network IP
        :param user: User configured on the devices
        :param password: Password for this user
        """
        self.eumel_api_url = ip + '/wizard/public/api/rest'
        self.auth = (user, password)
        super().__init__(manufacturer='Verbund', model='Eumel v1', serial_number=None, energy_unit=EnergyUnit.WATT_HOUR,
                         is_accumulated=True)

    def read_state(self, path=None) -> EnergyData:
        if path:
            tree = ElementTree.parse(path)
            with open(path) as file:
                raw = file.read()
        else:
            http_packet = requests.get(self.eumel_api_url, auth=self.auth)
            raw = http_packet.content.decode()
            tree = ElementTree.ElementTree(ElementTree.fromstring(raw))
        tree_root = tree.getroot()
        tree_header = tree_root[0].attrib
        tree_leaves = {child.attrib['id']: child.text for child in tree_root[0][1]}
        device = EnergyDevice(
            manufacturer=tree_header['man'],
            model=tree_header['mod'],
            serial_number=tree_header['sn'],
            energy_unit=self.energy_unit,
            is_accumulated=self.is_accumulated)
        access_epoch = int(time.time())
        time_format = '%Y-%m-%dT%H:%M:%SZ'
        energy = float(tree_leaves['TotWhImp'])
        mwh_energy = self.to_mwh(energy, self.energy_unit)
        measurement_epoch = int(time.mktime(time.strptime(tree_header['t'], time_format)))
        return EnergyData(asset=device, access_epoch=access_epoch, raw=raw, measurement_epoch=measurement_epoch,
                          energy=mwh_energy)

    def write_state(self, *args, **kwargs):
        raise NotImplementedError



"""
Library containing the implementations of CO2 oracles integration classes
"""
import calendar
import datetime
import requests

from energyweb.interfaces import ExternalData, IntegrationPoint


class CarbonEmissionData(ExternalData):
    """
    Standard for collected carbon emission data, to be transformed into mintable data.
    """

    def __init__(self, access_epoch, raw, accumulated_co2, measurement_epoch):
        """
        Minimum set of data for power measurement logging.
        :param access_epoch: Time the external API was accessed
        :param raw: Raw data collected
        :param accumulated_co2: Registered in kg of carbon dioxide
        :param measurement_epoch: Time of measurement at the source
        """
        self.accumulated_co2 = accumulated_co2
        self.measurement_epoch = measurement_epoch
        ExternalData.__init__(self, access_epoch, raw)


class CarbonEmissionAPI(IntegrationPoint):
    """
    Carbon emission endpoint data interface
    """
    def read_state(self, *args, **kwargs) -> CarbonEmissionData:
        """
        Establishes a connection to the integration medium and returns the latest state
        :rtype: ExternalData
        """
        raise NotImplementedError

    def write_state(self, *args, **kwargs) -> ExternalData:
        """
        Establishes a connection to the integration medium and persists data
        :rtype: ExternalData
        """
        raise NotImplementedError


class WattimeV1(CarbonEmissionAPI):

    def __init__(self, usr: str, pwd: str, ba: str, hours_from_now: int = 2):
        """
        Wattime API credentials. http://watttime.org/
        :param usr: Username used for login
        :param pwd: Users password
        :param ba: Balancing Authority. https://api.watttime.org/tutorials/#ba
        :param hours_from_now: Hours from the current time to check for CO emission. If none provided, will \
        get current day.
        """
        self.credentials = {'username': usr, 'password': pwd}
        self.api_url = 'https://api.watttime.org/api/v1/'
        self.ba = ba
        self.hours_from_now = hours_from_now

    def read_state(self) -> CarbonEmissionData:
        """
        Reach wattime api, parse and convert to CarbonEmissionData.
        """
        auth_token = self.__get_auth_token()
        # 2. Fetch marginal data
        raw = self.__get_marginal(auth_token)
        # 3. Converts lb/MW to kg/W
        accumulated_co2 = raw['marginal_carbon']['value'] * 0.453592 * pow(10, -6)
        # 4. Converts time stamps to epoch
        now = datetime.datetime.now()
        access_epoch = calendar.timegm(now.timetuple())
        measurement_timestamp = datetime.datetime.strptime(raw['timestamp'], "%Y-%m-%dT%H:%M:%SZ")
        measurement_epoch = calendar.timegm(measurement_timestamp.timetuple())
        return CarbonEmissionData(access_epoch, raw, accumulated_co2, measurement_epoch)

    def write_state(self, *args, **kwargs) -> ExternalData:
        raise NotImplementedError

    def __get_auth_token(self) -> str:
        """
        Exchange credentials for an access token.
        :return: Access token string suitable for passing as arg in other methods.
        """
        endpoint = self.api_url + 'obtain-token-auth/'
        r = requests.post(endpoint, data=self.credentials)
        if not r.status_code == 200:
            raise AttributeError('Failed getting a new token.')
        ans = r.json()
        if len(ans['token']) < 5:
            raise AttributeError('Failed getting a new token.')
        return ans['token']

    def __get_marginal(self, auth_token: str) -> dict:
        """
        Gets marginal carbon emission based on real time energy source mix of the grid.
        :param auth_token: authentication token
        :return: Measured data in lb/MW plus other relevant raw metadata.
        """
        base_time = datetime.datetime.now()
        if self.hours_from_now:
            start_time = base_time - datetime.timedelta(hours=self.hours_from_now)
            start_at = start_time.strftime("%Y-%m-%dT%H:00:00")
            end_at = base_time.strftime("%Y-%m-%dT%H:%M:%S")
        else:
            start_at = base_time.strftime("%Y-%m-%dT00:00:00")
            end_at = base_time.strftime("%Y-%m-%dT23:59:59")

        marginal_query = {
            'ba': self.ba,
            'start_at': start_at,
            'end_at': end_at,
            'page_size': 1,
            'market': 'RTHR'
        }
        endpoint = self.api_url + 'marginal/'
        h = {'Authorization': 'Token ' + auth_token}
        r = requests.get(endpoint, headers=h, params=marginal_query)
        ans = r.json()
        if 'count' not in ans.keys() and 'detail' in ans.keys():
            raise AttributeError('Failed to login on api.')
        if ans['count'] < 1:
            raise AttributeError('Empty response from api.')
        return ans['results'][0]

    def get_ba(self, lon, lat, auth_token) -> str:
        """
        Fetch Balancing Authority data based on geo spatial coordinates.
        :param lon: longitude
        :param lat: latitude
        :param auth_token: authentication token
        :return: Abbreviated ba name suitable for marginal requests.
        """
        geo_query = {
            'type': 'Point',
            'coordinates': [lon, lat]
        }
        endpoint = self.api_url + 'balancing_authorities/'
        h = {'token': auth_token}
        r = requests.get(endpoint, headers=h, params=geo_query)
        ans = r.json()
        return ans['abbrev']


class WattimeV2(CarbonEmissionAPI):

    def __init__(self, usr: str, pwd: str, ba: str):
        """
        Wattime API credentials. http://watttime.org/
        :param usr: Username used for login
        :param pwd: Users password
        :param ba: Balancing Authority. https://api.watttime.org/tutorials/#ba
        """
        self.credentials = (usr, pwd)
        self.api_url = 'https://api2.watttime.org/v2test/'
        self.ba = ba

    def read_state(self) -> CarbonEmissionData:
        """
        Reach wattime api, parse and convert to CarbonEmissionData.
        """
        auth_token = self.__get_auth_token()
        # 2. Fetch marginal data
        raw = self.__get_marginal(auth_token)
        # 3. Converts lb/MW to kg/W
        accumulated_co2 = raw['avg'] * 0.453592 * pow(10, -6)
        # 4. Converts time stamps to epoch
        now = datetime.datetime.now()
        access_epoch = calendar.timegm(now.timetuple())
        measurement_timestamp = now
        measurement_epoch = calendar.timegm(measurement_timestamp.timetuple())
        return CarbonEmissionData(access_epoch, raw, accumulated_co2, measurement_epoch)

    def write_state(self, *args, **kwargs) -> ExternalData:
        raise NotImplementedError

    def __get_auth_token(self) -> str:
        """
        Exchange credentials for an access token.
        :return: Access token string suitable for passing as arg in other methods.
        """
        endpoint = self.api_url + 'login'
        r = requests.get(endpoint, auth=self.credentials)
        if not r.status_code == 200:
            raise AttributeError('Failed getting a new token.')
        ans = r.json()
        if len(ans['token']) < 5:
            raise AttributeError('Failed getting a new token.')
        return ans['token']

    def __get_marginal(self, auth_token: str) -> dict:
        """
        Gets marginal carbon emission based on real time energy source mix of the grid.
        :param auth_token: authentication token
        :return: Measured data in lb/MW plus other relevant raw metadata.
        """
        marginal_query = {
            'ba': self.ba
        }
        endpoint = self.api_url + 'insight/'
        h = {'Authorization': 'Bearer ' + auth_token}
        r = requests.get(endpoint, headers=h, params=marginal_query)
        if not r.status_code == 200:
            raise AttributeError('Failed to login on api.')
        return r.json()



"""
Asset Registry Abstract Smart-Contract interface
"""
contract = {
    "address": "0x",
    "abi": [
        {'constant': True, 'inputs': [], 'name': 'db', 'outputs': [{'name': '', 'type': 'address'}], 'payable': False,
         'stateMutability': 'view', 'type': 'function'},
        {'constant': True, 'inputs': [], 'name': 'cooContract', 'outputs': [{'name': '', 'type': 'address'}],
         'payable': False, 'stateMutability': 'view', 'type': 'function'},
        {'constant': True, 'inputs': [{'name': '_role', 'type': 'uint8'}, {'name': '_caller', 'type': 'address'}],
         'name': 'isRole', 'outputs': [{'name': '', 'type': 'bool'}], 'payable': False, 'stateMutability': 'view',
         'type': 'function'}, {'anonymous': False, 'inputs': [{'indexed': False, 'name': 'sender', 'type': 'address'},
                                                              {'indexed': True, 'name': '_assetId', 'type': 'uint256'}],
                               'name': 'LogAssetCreated', 'type': 'event'},
        {'anonymous': False, 'inputs': [{'indexed': True, 'name': '_assetId', 'type': 'uint256'}],
         'name': 'LogAssetFullyInitialized', 'type': 'event'},
        {'anonymous': False, 'inputs': [{'indexed': True, 'name': '_assetId', 'type': 'uint256'}],
         'name': 'LogAssetSetActive', 'type': 'event'},
        {'anonymous': False, 'inputs': [{'indexed': True, 'name': '_assetId', 'type': 'uint256'}],
         'name': 'LogAssetSetInactive', 'type': 'event'},
        {'constant': False, 'inputs': [], 'name': 'createAsset', 'outputs': [], 'payable': False,
         'stateMutability': 'nonpayable', 'type': 'function'},
        {'constant': False, 'inputs': [{'name': '_dbAddress', 'type': 'address'}], 'name': 'init', 'outputs': [],
         'payable': False, 'stateMutability': 'nonpayable', 'type': 'function'}, {'constant': False, 'inputs': [
            {'name': '_assetId', 'type': 'uint256'}, {'name': '_country', 'type': 'bytes32'},
            {'name': '_region', 'type': 'bytes32'}, {'name': '_zip', 'type': 'bytes32'},
            {'name': '_city', 'type': 'bytes32'}, {'name': '_street', 'type': 'bytes32'},
            {'name': '_houseNumber', 'type': 'bytes32'}, {'name': '_gpsLatitude', 'type': 'bytes32'},
            {'name': '_gpsLongitude', 'type': 'bytes32'}], 'name': 'initLocation', 'outputs': [], 'payable': False,
                                                                                  'stateMutability': 'nonpayable',
                                                                                  'type': 'function'},
        {'constant': False, 'inputs': [{'name': '_assetId', 'type': 'uint256'}, {'name': '_active', 'type': 'bool'}],
         'name': 'setActive', 'outputs': [], 'payable': False, 'stateMutability': 'nonpayable', 'type': 'function'},
        {'constant': False, 'inputs': [{'name': '_newLogic', 'type': 'address'}], 'name': 'update', 'outputs': [],
         'payable': False, 'stateMutability': 'nonpayable', 'type': 'function'},
        {'constant': True, 'inputs': [{'name': '_assetId', 'type': 'uint256'}], 'name': 'getActive',
         'outputs': [{'name': '', 'type': 'bool'}], 'payable': False, 'stateMutability': 'view', 'type': 'function'},
        {'constant': True, 'inputs': [{'name': '_assetId', 'type': 'uint256'}], 'name': 'getLastSmartMeterReadFileHash',
         'outputs': [{'name': 'datalog', 'type': 'bytes32'}], 'payable': False, 'stateMutability': 'view',
         'type': 'function'},
        {'constant': True, 'inputs': [], 'name': 'getAssetListLength', 'outputs': [{'name': '', 'type': 'uint256'}],
         'payable': False, 'stateMutability': 'view', 'type': 'function'},
        {'constant': True, 'inputs': [{'name': '_assetId', 'type': 'uint256'}], 'name': 'getAssetLocation',
         'outputs': [{'name': 'country', 'type': 'bytes32'}, {'name': 'region', 'type': 'bytes32'},
                     {'name': 'zip', 'type': 'bytes32'}, {'name': 'city', 'type': 'bytes32'},
                     {'name': 'street', 'type': 'bytes32'}, {'name': 'houseNumber', 'type': 'bytes32'},
                     {'name': 'gpsLatitude', 'type': 'bytes32'}, {'name': 'gpsLongitude', 'type': 'bytes32'}],
         'payable': False, 'stateMutability': 'view', 'type': 'function'}, {'constant': False, 'inputs': [
            {'name': '_assetId', 'type': 'uint256'}, {'name': '_newSmartMeter', 'type': 'address'}],
                                                                            'name': 'updateSmartMeter', 'outputs': [],
                                                                            'payable': False,
                                                                            'stateMutability': 'nonpayable',
                                                                            'type': 'function'}],
    "bytecode": '0x'
}



"""
Asynchronous event watcher loop
"""
import asyncio
import datetime


class Task:
    """
    Tasks are routines that run from time to time respecting an interval and spawn coroutines.
    These routines may only execute if a trigger condition is fired.
    """
    def __init__(self, queue: {str: asyncio.Queue}, polling_interval: datetime.timedelta = None, eager: bool = False, run_forever: bool = True):
        """
        :param polling_interval: in seconds
        :param queue: app asyncio queues for messages exchange between threads
        :param eager: if main waits polling time first or is eager to start
        """
        self.polling_interval = polling_interval
        self.queue = queue
        self.run_forever = run_forever
        self.eager = eager

    async def _prepare(self):
        """
        Perform steps required prior to running the main task
        run exactly once
        """
        raise NotImplementedError

    async def _main(self, *args):
        """
        The main task
        """
        raise NotImplementedError

    async def _finish(self):
        """
        Perform steps required after running the main task
        run exactly once
        """
        raise NotImplementedError

    def _handle_exception(self, e: Exception):
        """
        Handle exceptions when they occur
        """
        raise NotImplementedError

    async def run(self, *args):
        """
        run all steps of the task
        """
        async def main_loop():
            if self.polling_interval and not self.eager:
                await asyncio.sleep(self.polling_interval.total_seconds())
            await self._main(*args)
            if self.polling_interval and self.eager:
                await asyncio.sleep(self.polling_interval.total_seconds())
        try:
            await self._prepare()
            try:
                await main_loop()
                while self.run_forever:
                    await asyncio.ensure_future(main_loop())
            except Exception as e:
                self._handle_exception(e)
            finally:
                await self._finish()
                if self.run_forever:
                    await self.run(*args)
        except Exception as e:
            self._handle_exception(e)


class App:
    """
    General application abstraction
    """
    def __init__(self):
        self.tasks: [asyncio.tasks] = []
        self.queue: {str: asyncio.Queue} = {}
        self.loop = asyncio.get_event_loop()
        self._configure()

    def _configure(self):
        """
        Configuration phase of the app.
        """
        raise NotImplementedError

    def _clean_up(self):
        """
        Overwrite this method to implement clean up logic like closing sockets and streams.
        """
        raise NotImplementedError

    def _handle_exception(self, e: Exception):
        """
        Handle exceptions when they occur
        """
        raise NotImplementedError

    def _register_task(self, task: Task, *args):
        """
        Add task to be executed in run time
        """
        if not task:
            raise Exception('Please add a Task type with callable task named method.')
        self.tasks.append((task, args))

    def _register_queue(self, queue_id: str, max_size: int = 0):
        """
        Creates a new asyncio Queue for messages exchange between threads
        :param queue_id: Name of queue index
        :param max_size: Maximum number of messages
        """
        self.queue[queue_id] = asyncio.Queue(maxsize=max_size)

    def run(self):
        """
        execute all tasks in task list in their own thread
        """
        try:
            self.loop.run_until_complete(asyncio.gather(*[task.run(*args) for task, args in self.tasks]))
        except Exception as e:
            self._handle_exception(e)
        finally:
            self._clean_up()



"""
Library containing the Certificate of Origin v1.0 integration classes
"""
from energyweb.eds.interfaces import EnergyData
from energyweb.smart_contract.interfaces import EVMSmartContractClient
from energyweb.smart_contract.origin.consumer_v1 import contract as consumer_v1
from energyweb.smart_contract.origin.producer_v1 import contract as producer_v1
from energyweb.smart_contract.origin.asset_reg_v1 import contract as asset_reg_v1


class ProducedEnergy(EnergyData):
    """
    Helper for coo smart-contract mint_produced method
    """
    def __init__(self, value: int, is_meter_down: bool, previous_hash: str, co2_saved: int, is_co2_down: bool):
        """
        :type previous_hash: previous
        :param value:  Time the value was measured in epoch format
        :param is_meter_down:  Measured value
        """
        self.value = value
        self.is_meter_down = is_meter_down
        self.previous_hash = previous_hash
        self.co2_saved = co2_saved
        self.is_co2_down = is_co2_down


class ConsumedEnergy(EnergyData):
    """
    Helper for coo smart-contract mint_consumed method
    """
    def __init__(self, value: int, previous_hash: str, is_meter_down: bool):
        """
        :type previous_hash: previous
        :param value:  Time the value was measured in epoch format
        :param is_meter_down:  Measured value
        """
        self.value = value
        self.is_meter_down = is_meter_down
        self.previous_hash = previous_hash


class OriginV1(EVMSmartContractClient):
    """
    Origin in as smart-contract system deployed on Energy Web Blockchain network.
    It is designed to issue and validate green energy certificates of origin.

    This class is only an interface to a ewf-client via json rpc calls and interact with the smart-contract.
    """

    def mint(self, energy: EnergyData) -> dict:
        raise NotImplementedError

    def __init__(self, asset_id: int, wallet_add: str, wallet_pwd: str, client_url: str):
        """
        :param asset_id: ID received in device registration.
        :param wallet_add: Network wallet address
        :param wallet_add: Network wallet password
        :param client_url: URL like address to the blockchain client api.
        contract_address is not used from task_config
        """
        contracts = {
            "producer": producer_v1,
            "consumer": consumer_v1,
            "asset_reg": asset_reg_v1
        }
        credentials = (wallet_add, wallet_pwd)
        max_retries = 1000
        retry_pause = 5

        self.asset_id = asset_id
        super().__init__(credentials, contracts, client_url, max_retries, retry_pause)

    def register_asset(self, country: str, region: str, zip_code: str, city: str, street: str, house_number: str,
                       latitude: str, longitude: str):
        """
        Register device. The account signing the transaction must have "AssetAdmin" role to successfully register.
        The device registry is done manually on this project phase. Please contact the EWF's Ramp up team.

        Source:
            AssetLogic.sol
        Call stack:
            function createAsset() external onlyRole(RoleManagement.Role.AssetAdmin)
            function initLocation ( uint _index, bytes32 _country, bytes32 _region, bytes32 _zip, bytes32 _city, bytes32 _street, bytes32 _houseNumber, bytes32 _gpsLatitude, bytes32 _gpsLongitude ) external isInitialized onlyRole(RoleManagement.Role.AssetAdmin)
            function initGeneral ( uint _index, address _smartMeter, address _owner, AssetType _assetType, Compliance _compliance, uint _operationalSince, uint _capacityWh, bool _active ) external isInitialized userHasRole(RoleManagement.Role.AssetManager, _owner) onlyRole(RoleManagement.Role.AssetAdmin)
        Wait for:
            event LogAssetCreated(address sender, uint id);
            event LogAssetFullyInitialized(uint id);
        TODO: Implement in next releases
        """
        pass


class OriginProducer(OriginV1):
    """
    Green Energy Producer

    Origin in as smart-contract system deployed on Energy Web Blockchain network.
    It is designed to issue and validate green energy certificates of origin.

    This class is only an interface to a ewf-client via json rpc calls and interact with the smart-contract.
    """

    def mint(self, energy: ProducedEnergy) -> dict:
        """
        Source:
            AssetProducingRegistryLogic.sol
        Call stack:
            function saveSmartMeterRead( uint _assetId, uint _newMeterRead, bool _smartMeterDown, bytes32 _lastSmartMeterReadFileHash, uint _CO2OffsetMeterRead, bool _CO2OffsetServiceDown ) external isInitialized onlyAccount(AssetProducingRegistryDB(address(db)).getSmartMeter(_assetId))
        Wait for:
            event LogNewMeterRead(uint indexed _assetId, uint _oldMeterRead, uint _newMeterRead, bool _smartMeterDown, uint _certificatesCreatedForWh, uint _oldCO2OffsetReading, uint _newCO2OffsetReading, bool _serviceDown);
        """
        if not isinstance(energy.value, int):
            raise ValueError('No Produced energy present or in wrong format.')
        if not isinstance(energy.is_meter_down, bool):
            raise ValueError('No Produced energy status present or in wrong format.')
        if not isinstance(energy.previous_hash, str):
            raise ValueError('No Produced hash of last file present or in wrong format.')
        if not isinstance(energy.co2_saved, int):
            raise ValueError('No Produced co2 present or in wrong format.')
        if not isinstance(energy.is_co2_down, bool):
            raise ValueError('No Produced co2 status present or in wrong format.')

        receipt = self.send_raw('producer', 'saveSmartMeterRead', self.asset_id, energy.value,
                                energy.is_meter_down, energy.previous_hash.encode(),
                                energy.co2_saved, energy.is_co2_down)
        if not receipt:
            raise ConnectionError
        return receipt

    def last_hash(self):
        """
        Get last file hash registered from producer contract
        Source:
            AssetLogic.sol
        Call stack:
            function getAssetDataLog(uint _assetId)
        """
        receipt = self.call('producer', 'getLastSmartMeterReadFileHash', self.asset_id)
        if not receipt:
            raise ConnectionError
        return receipt

    def last_state(self):
        """
        Get complete state of initiated Asset
        Source:
            AssetLogic.sol
        Call stack:
            function getAssetGeneral(uint _assetId)
                external
                constant
                returns(
                    address _smartMeter,
                    address _owner,
                    uint _operationalSince,
                    uint _lastSmartMeterReadWh,
                    bool _active,
                    bytes32 _lastSmartMeterReadFileHash
                    )
        """
        # TODO store with names keys
        receipt = self.call('producer', 'getAssetGeneral', self.asset_id)
        if not receipt:
            raise ConnectionError
        return receipt


class OriginConsumer(OriginV1):
    """
    Green Energy Consumer

    Origin in as smart-contract system deployed on Energy Web Blockchain network.
    It is designed to issue and validate green energy certificates of origin.

    This class is only an interface to a ewf-client via json rpc calls and interact with the smart-contract.
    """

    def mint(self, energy: ConsumedEnergy) -> dict:
        """
        Source:
            AssetConsumingRegistryLogic.sol
        Call stack:
            function saveSmartMeterRead(uint _assetId, uint _newMeterRead, bytes32 _lastSmartMeterReadFileHash, bool _smartMeterDown) external isInitialized onlyAccount(AssetConsumingRegistryDB(address(db)).getSmartMeter(_assetId))
        Wait for:
            event LogNewMeterRead(uint indexed _assetId, uint _oldMeterRead, uint _newMeterRead, uint _certificatesUsedForWh, bool _smartMeterDown);
        """
        if not isinstance(energy.value, int):
            raise ValueError('No Produced energy present or in wrong format.')
        if not isinstance(energy.is_meter_down, bool):
            raise ValueError('No Produced energy status present or in wrong format.')
        if not isinstance(energy.previous_hash, str):
            raise ValueError('No Produced hash of last file present or in wrong format.')
        receipt = self.send_raw('consumer', 'saveSmartMeterRead', self.asset_id, energy.value,
                                energy.previous_hash.encode(), energy.is_meter_down)
        if not receipt:
            raise ConnectionError
        return receipt

    def last_hash(self):
        """
        Get last file hash registered from consumer contract
        Source:
            AssetLogic.sol
        Call stack:
            function getAssetDataLog(uint _assetId)
        """
        receipt = self.call('consumer', 'getLastSmartMeterReadFileHash', self.asset_id)
        if not receipt:
            raise ConnectionError
        return receipt

    def last_state(self):
        """
        Get last file hash registered from producer contract
        Source:
            AssetLogic.sol
        Call stack:
            function getAssetGeneral(uint _assetId)
                external
                constant
                returns (
                    address _smartMeter,
                    address _owner,
                    uint _operationalSince,
                    uint _capacityWh,
                    bool _maxCapacitySet,
                    uint _lastSmartMeterReadWh,
                    uint _certificatesUsedForWh,
                    bool _active,
                    bytes32 _lastSmartMeterReadFileHash
                    )
        """
        receipt = self.call('consumer', 'getAssetGeneral', self.asset_id)
        if not receipt:
            raise ConnectionError
        return receipt



import time

from web3 import Web3, HTTPProvider
from web3.contract import ConciseContract
from web3.utils.filters import Filter

from energyweb.eds.interfaces import EnergyData
from energyweb.carbonemission import CarbonEmissionData
from energyweb.interfaces import BlockchainClient


class GreenEnergy(EnergyData, CarbonEmissionData):
    """
    Green energy data read from external data sources of energy and carbon emissions
    """

    def __init__(self, energy: EnergyData, carbon_emission: CarbonEmissionData):
        self.energy = energy
        self.carbon_emission = carbon_emission
        super().__init__(None, None)

    def co2_saved(self, energy_offset: int = 0):
        """
        Returns amount of carbon dioxide saved
        :param energy_offset: Energy already registered to add up in case the energy measured was not accumulated
        :return: Total amount of kg of carbon dioxide
        """
        if not self.carbon_emission:
            return 0
        accumulated_energy = self.energy.energy
        if not self.energy.device.is_accumulated:
            accumulated_energy += energy_offset
        co2_saved = self.carbon_emission.accumulated_co2
        calculated_co2 = accumulated_energy * co2_saved
        co2_saved = int(calculated_co2 * pow(10, 3))
        return co2_saved


class EVMSmartContractClient(BlockchainClient):
    """
    General EVM based blockchain client smart contract integration.
    Tested on:
        - https://github.com/paritytech/parity
        - https://github.com/energywebfoundation/energyweb-client
    """

    def __init__(self, credentials: tuple, contracts: dict, client_url: str, max_retries: int, retry_pause: int):
        """
        :param credentials: Network credentials ( address, password )
        :param contracts: Contracts structure containing abi and bytecode keys.
        :param client_url: URL like address to the blockchain client api.
        :param max_retries: Software will try to connect to provider this amount of times
        :param retry_pause: Software will wait between reconnection trials this amount of seconds
        """
        self.MAX_RETRIES = max_retries
        self.SECONDS_BETWEEN_RETRIES = retry_pause
        self.w3 = Web3(HTTPProvider(client_url))
        self.credentials = credentials
        self.contracts = contracts

    def is_synced(self) -> bool:
        """
        Simple algorithm to check if the blockchain client is synced to the latest block.
        :return: Is synced true or false
        """
        synced_block = str(self.w3.eth.blockNumber)
        latest_block_obj = self.w3.eth.getBlock('latest')
        latest_block = str(latest_block_obj.number)
        return synced_block == latest_block

    def send(self, contract_name: str, method_name: str, *args) -> dict:
        """
        Sends a regular transaction to call a smart-contract method. This requires the user have the keys previously
        imported to the blockchain client and to accept the transaction on Metamask or on the client's user interface.
        :param contract_name: Contract key as in the contracts list used to instantiate this class.
        :param method_name: Method name as in the contract abi.
        :param args: Arguments passed when calling the method. Must be in the same order as in the abi.
        :return: The transaction receipt after mining is confirmed.
        """
        # TODO: Check for more elegant way of verifying the tx receipt
        if not self.is_synced():
            raise ConnectionError('Client is not synced to the last block.')
        self.w3.personal.unlockAccount(account=self.w3.toChecksumAddress(self.credentials[0]),
                                       passphrase=self.credentials[1])
        contract = self.contracts[contract_name]
        contract_instance = self.w3.eth.contract(
            abi=contract['abi'],
            address=self.w3.toChecksumAddress(contract['address']),
            bytecode=contract['bytecode'],
            ContractFactoryClass=ConciseContract)
        tx_hash = getattr(contract_instance, method_name)(*args, transact={
            'from': self.w3.toChecksumAddress(self.credentials[0])})
        if not tx_hash:
            raise ConnectionError('Transaction was not sent.')
        tx_receipt = None
        for _ in range(self.MAX_RETRIES):
            tx_receipt = self.w3.eth.getTransactionReceipt(tx_hash)
            if tx_receipt and tx_receipt['blockNumber']:
                break
            time.sleep(self.SECONDS_BETWEEN_RETRIES)
        return tx_receipt

    def call(self, contract_name: str, method_name: str, *args) -> dict:
        """
        Calls a smart-contract method without sending a transaction. Suitable for read-only operations.
        :param contract_name: Contract key as in the contracts list used to instantiate this class.
        :param method_name: Method name as in the contract abi.
        :param args: Arguments passed when calling the method. Must be in the same order as in the abi.
        :return: The transaction receipt after mining is confirmed.
        """
        if not self.is_synced():
            raise ConnectionError('Client is not synced to the last block.')
        contract = self.contracts[contract_name]
        contract_instance = self.w3.eth.contract(
            abi=contract['abi'],
            address=self.w3.toChecksumAddress(contract['address']),
            bytecode=contract['bytecode'],
            ContractFactoryClass=ConciseContract)
        return getattr(contract_instance, method_name)(*args)

    def send_raw(self, contract_name: str, method_name: str, *args) -> dict:
        """
        Sends a raw transaction to call a smart-contract method.
        First it creates the transaction, then fetches the account tx count - to avoid repetition attacks,
        determines gas price, signs it, and finally sends it to the blockchain client.
        :param contract_name: Contract key as in the contracts list used to instantiate this class.
        :param method_name: Method name as in the contract abi.
        :param args: Arguments passed when calling the method. Must be in the same order as in the abi.
        :return: The transaction receipt after mining is confirmed.
        """
        contract = self.contracts[contract_name]
        contract_instance = self.w3.eth.contract(
            abi=contract['abi'],
            address=self.w3.toChecksumAddress(contract['address']),
            bytecode=contract['bytecode'])

        if not self.is_synced():
            raise ConnectionError('Client is not synced to the last block.')

        nonce = self.w3.eth.getTransactionCount(account=self.w3.toChecksumAddress(self.credentials[0]))
        transaction = {
            'from': self.w3.toChecksumAddress(self.credentials[0]),
            'gas': 400000,
            'gasPrice': self.w3.toWei('0', 'gwei'),
            'nonce': nonce,
        }
        tx = getattr(contract_instance.functions, method_name)(*args).buildTransaction(transaction)
        private_key = bytearray.fromhex(self.credentials[1])
        signed_txn = self.w3.eth.account.signTransaction(tx, private_key=private_key)
        tx_hash = self.w3.eth.sendRawTransaction(signed_txn.rawTransaction)

        if not tx_hash:
            raise ConnectionError('Transaction was not sent.')
        tx_receipt = None
        for _ in range(self.MAX_RETRIES):
            tx_receipt = self.w3.eth.getTransactionReceipt(tx_hash)
            if tx_receipt and tx_receipt['blockNumber']:
                break
            time.sleep(self.SECONDS_BETWEEN_RETRIES)
        return tx_receipt

    def create_event_filter(self, contract_name: str, event_name: str, block_count: int = 1000) -> Filter:
        """
        Create Filter on the client, the client must have the option enabled or it might fail.
        :param contract_name: Contract key as in the contracts list used to instantiate this class.
        :param event_name: Like written in the abi
        :param block_count: Number of blocks prior to the latest to start filtering from
        :return: Filter
        """
        contract = self.contracts[contract_name]
        contract_instance = self.w3.eth.contract(
            abi=contract['abi'],
            address=self.w3.toChecksumAddress(contract['address']),
            bytecode=contract['bytecode'])
        latest_block = self.w3.eth.getBlock('latest')
        return getattr(contract_instance.events, event_name)().createFilter(fromBlock=latest_block.number - block_count)

    def create_event_trigger(self, contract_name: str, event_name: str, block_count: int = 1000) -> dict:
        """
        Todo: Fix this to get the blocks and check for new events, demands memory or persistence
        Create Filter on the client, the client must have the option enabled or it might fail.
        :param contract_name: Contract key as in the contracts list used to instantiate this class.
        :param event_name: Like written in the abi
        :param block_count: Number of blocks prior to the latest to start filtering from
        :return: Filter
        """
        usn_filter = self.create_event_filter(contract_name=contract_name, event_name=event_name, block_count=block_count)
        event_logs = usn_filter.get_all_entries()
        # sessionId = Web3.toHex(event_logs[-1]['transactionHash'])
        return event_logs

    def mint(self, energy: EnergyData) -> dict:
        """
        Mint the measured energy in the blockchain smart-contract
        """
        raise NotImplementedError


"""
USN RentingSupport Smart-Contract interface
"""
contract = {
    "address": "0x85Ec283a3Ed4b66dF4da23656d4BF8A507383bca",
    "abi": [
        {
            "constant": True,
            "inputs": [
                {
                    "name": "id",
                    "type": "bytes32"
                },
                {
                    "name": "index",
                    "type": "uint256"
                }
            ],
            "name": "getState",
            "outputs": [
                {
                    "name": "controller",
                    "type": "address"
                },
                {
                    "name": "rentedFrom",
                    "type": "uint64"
                },
                {
                    "name": "rentedUntil",
                    "type": "uint64"
                },
                {
                    "name": "properties",
                    "type": "uint128"
                }
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function"
        },
        {
            "constant": True,
            "inputs": [
                {
                    "name": "id",
                    "type": "bytes32"
                }
            ],
            "name": "getStateCount",
            "outputs": [
                {
                    "name": "",
                    "type": "uint256"
                }
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function"
        },
        {
            "constant": False,
            "inputs": [
                {
                    "name": "id",
                    "type": "bytes32"
                },
                {
                    "name": "secondsToRent",
                    "type": "uint32"
                },
                {
                    "name": "token",
                    "type": "address"
                }
            ],
            "name": "rent",
            "outputs": [],
            "payable": True,
            "stateMutability": "payable",
            "type": "function"
        },
        {
            "constant": True,
            "inputs": [
                {
                    "name": "id",
                    "type": "bytes32"
                },
                {
                    "name": "user",
                    "type": "address"
                }
            ],
            "name": "getRentingState",
            "outputs": [
                {
                    "name": "rentable",
                    "type": "bool"
                },
                {
                    "name": "free",
                    "type": "bool"
                },
                {
                    "name": "open",
                    "type": "bool"
                },
                {
                    "name": "controller",
                    "type": "address"
                },
                {
                    "name": "rentedUntil",
                    "type": "uint64"
                },
                {
                    "name": "rentedFrom",
                    "type": "uint64"
                },
                {
                    "name": "props",
                    "type": "uint128"
                }
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function"
        },
        {
            "constant": False,
            "inputs": [
                {
                    "name": "id",
                    "type": "bytes32"
                }
            ],
            "name": "returnObject",
            "outputs": [],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function"
        },
        {
            "constant": True,
            "inputs": [
                {
                    "name": "id",
                    "type": "bytes32"
                }
            ],
            "name": "supportedTokens",
            "outputs": [
                {
                    "name": "addresses",
                    "type": "address[]"
                }
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function"
        },
        {
            "constant": True,
            "inputs": [
                {
                    "name": "id",
                    "type": "bytes32"
                },
                {
                    "name": "token",
                    "type": "address"
                }
            ],
            "name": "tokenReceiver",
            "outputs": [
                {
                    "name": "",
                    "type": "bytes32"
                }
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function"
        },
        {
            "constant": True,
            "inputs": [
                {
                    "name": "id",
                    "type": "bytes32"
                },
                {
                    "name": "user",
                    "type": "address"
                },
                {
                    "name": "secondsToRent",
                    "type": "uint32"
                },
                {
                    "name": "token",
                    "type": "address"
                }
            ],
            "name": "price",
            "outputs": [
                {
                    "name": "",
                    "type": "uint128"
                }
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function"
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "name": "fixFilter",
                    "type": "bytes32"
                },
                {
                    "indexed": True,
                    "name": "id",
                    "type": "bytes32"
                },
                {
                    "indexed": False,
                    "name": "controller",
                    "type": "address"
                },
                {
                    "indexed": False,
                    "name": "rentedFrom",
                    "type": "uint64"
                },
                {
                    "indexed": False,
                    "name": "rentedUntil",
                    "type": "uint64"
                },
                {
                    "indexed": False,
                    "name": "noReturn",
                    "type": "bool"
                },
                {
                    "indexed": False,
                    "name": "amount",
                    "type": "uint128"
                },
                {
                    "indexed": False,
                    "name": "token",
                    "type": "address"
                },
                {
                    "indexed": False,
                    "name": "properties",
                    "type": "uint128"
                }
            ],
            "name": "LogRented",
            "type": "event"
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "name": "fixFilter",
                    "type": "bytes32"
                },
                {
                    "indexed": True,
                    "name": "id",
                    "type": "bytes32"
                },
                {
                    "indexed": False,
                    "name": "controller",
                    "type": "address"
                },
                {
                    "indexed": False,
                    "name": "rentedFrom",
                    "type": "uint64"
                },
                {
                    "indexed": False,
                    "name": "rentedUntil",
                    "type": "uint64"
                },
                {
                    "indexed": False,
                    "name": "paidBack",
                    "type": "uint128"
                }
            ],
            "name": "LogReturned",
            "type": "event"
        }
    ],
    "bytecode": '0x'
}



import os
import logging
import colorlog


class Logger:
    """
    Instantiate loggers for general purpose
    """

    def __init__(self, log_name: str, store: str = None, enable_debug: bool = False):
        """
        :param log_name: File name for the log. Will create log entries and file name ie. my_app.log
        :param store: Path to folder where the log files will be stored in disk. Please note that disk can get full and prevent OS boot.
        :param enable_debug: Enabling debug creates a log for errors. Needs storage. Please manually delete it.
        """

        tty_handler = colorlog.StreamHandler()
        tty_handler.setFormatter(colorlog.ColoredFormatter('%(log_color)s%(message)s'))

        console = colorlog.getLogger(log_name)
        formatter = logging.Formatter('%(asctime)s [%(name)s][%(process)d][%(levelname)s]%(message)s')
        if store:
            if not os.path.exists(store):
                os.makedirs(store)
            file_handler = logging.FileHandler(os.path.join(store, f'{log_name}.log'))
            file_handler.setFormatter(formatter)
            console.addHandler(file_handler)
        console.addHandler(tty_handler)
        console.setLevel(logging.DEBUG)
        error_log = None
        if enable_debug:
            if not store:
                raise Exception('Needs storage path in store parameter defined.')
            error_log = logging.getLogger(log_name + '.error')
            error_file_handler = logging.FileHandler(os.path.join(store, f'{log_name}.error.log'))
            error_file_handler.setFormatter(formatter)
            error_log.addHandler(error_file_handler)
            error_log.setLevel(logging.WARNING)

        self.store = store
        self.console = console
        self.error_log = error_log
        self.enable_debug = enable_debug



"""
https://pypi.org/
https://pypi.org/classifiers/
"""
import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setuptools.setup(
    name="energyweb",
    version="0.4.0",
    author="github.com/cerealkill",
    author_email="paul.depraz@energyweb.org",
    description="Energy utility data interface for blockchain smart contracts",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/energywebfoundation/ew-link-bond",
    packages=setuptools.find_packages(exclude=["docs", "tests"]),
    install_requires=['web3>=4.8.0,<5.0.0', 'colorlog>=3.1.4', 'base58>=1.0.3'],
    keywords=['ethereum', 'blockchain', 'energy-web', 'energy', 'smart-energy_meter'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
        "Environment :: Console",
        "Development Status :: 2 - Pre-Alpha"
    ],
)



"""
Interface for the Bond API
- Bond API delivers prosumer data
"""
import calendar
import datetime
import json
import requests

from energyweb.eds.interfaces import EnergyDevice, EnergyData


class BondAPIv1(EnergyDevice):

    def __init__(self, base_url, source, device_id, user=None, password=None):
        """
        Bond api spec v1.x module

        :param base_url: Must start with protocol. ie. https://my-url
        :param source: Must be 'produced' or 'consumed'
        :param device_id: Please mind it will be url encoded.
        :param user: User credential.
        :param password: Password for basic authentication method.
        """
        if source not in ('produced', 'consumed'):
            raise AssertionError
        self.base_url = base_url
        self.api_url = '{}/{}/{}'.format(base_url, source, device_id)
        self.auth = (user, password)
        super().__init__(
            manufacturer='Slock.it',
            model='Virtual Energy Meter',
            serial_number='0001000',
            energy_unit='KILOWATT_HOUR',
            is_accumulated=False)
        # TODO get from device metadata from API call

    def read_state(self, start=None, end=None) -> EnergyData:
        # raw
        raw, data, measurement_list = self._reach_source(self.api_url, start, end)
        # device
        device_meta = data[-1]['device']
        device = EnergyDevice(**device_meta)
        # accumulated energy in Wh
        if device.is_accumulated:
            energy = self.to_wh(measurement_list[-1]['energy'], device.energy_unit)
        else:
            energy = self.to_wh(sum(i['energy'] for i in measurement_list), device.energy_unit)
        # access_epoch
        now = datetime.datetime.now().astimezone()
        access_epoch = calendar.timegm(now.timetuple())
        #  measurement epoch
        measurement_time = datetime.datetime.strptime(measurement_list[-1]['measurement_time'], "%Y-%m-%dT%H:%M:%S%z")
        measurement_epoch = calendar.timegm(measurement_time.timetuple())
        return EnergyData(device=device, access_epoch=access_epoch, raw=raw, energy=energy,
                          measurement_epoch=measurement_epoch)

    def write_state(self, *args, **kwargs) -> EnergyData:
        raise NotImplementedError

    def _reach_source(self, url, start=None, end=None, have_next=False) -> (str, dict):
        marginal_query = None
        if not have_next:
            # calculate the current time and one hour back from there
            end_date = datetime.datetime.now(datetime.timezone.utc) if not end else end
            start_date = end_date - datetime.timedelta(hours=1, minutes=15) if not start else start
            marginal_query = {
                'start': start_date.astimezone().isoformat(),  # expects year-month-day
                'end': end_date.isoformat(),  # the hour of day
                'limit': 10
            }
        http_packet = requests.get(url=url, params=marginal_query, auth=self.auth)
        raw = http_packet.content.decode()
        if not http_packet.ok:
            raise EnvironmentError
        data = http_packet.json()
        return self._parse_source(raw, data)

    def _parse_source(self, raw: str, data: dict):
        measurement_list = data['measuredEnergy']
        if 'next' in data and data['next']:
            _, __, ___ = self._reach_source(self.base_url + data['next'], have_next=True)
            return tuple(x + y for x, y in zip(([raw], [data], measurement_list), (_, __, ___)))
        return [raw], [data], measurement_list


# TODO: Add Tox tests for this module
class BondAPIv1TestDevice1(BondAPIv1):
    """
    Data parsing test fixture
    """

    def _reach_source(self, url, start=None, end=None, have_next=False) -> (str, dict):
        raw = {
            "base_url/produced/0": '{"count": 0, "previous": null, "next": "/second", "device": {"manufacturer": "Siemens", "model": "ABC-123", "serial_number": "345345345", "latitude": "54.443567", "longitude": "-23.312543", "energy_unit": "kilowatt_hour", "is_accumulated": true }, "measuredEnergy": [{"energy": 100, "measurement_time": "2018-03-15T10:30:00+00:00"}, {"energy": 200, "measurement_time": "2018-03-15T12:30:00+00:00"}, {"energy": 250, "measurement_time": "2018-03-15T14:30:00+00:00"} ] }',
            "base_url/second": '{"count": 1, "previous": "first", "next": null, "device": {"manufacturer": "Siemens", "model": "ABC-123", "serial_number": "345345345", "latitude": "54.443567", "longitude": "-23.312543", "energy_unit": "kilowatt_hour", "is_accumulated": true }, "measuredEnergy": [{"energy": 390, "measurement_time": "2018-03-15T16:30:00+00:00"}, {"energy": 400, "measurement_time": "2018-03-15T18:30:00+00:00"} ] }'
        }
        data = json.loads(raw[url])
        return self._parse_source(str(raw), data)


class BondAPIv1TestDevice2(BondAPIv1):
    """
    Data parsing test fixture
    """

    def _reach_source(self, url, start=None, end=None, have_next=False) -> (str, dict):
        raw = {
            "base_url/produced/0": '{"count": 0, "previous": null, "next": "/second", "device": {"manufacturer": "Siemens", "model": "ABC-123", "serial_number": "345345345", "latitude": "54.443567", "longitude": "-23.312543", "energy_unit": "watt_hour", "is_accumulated": false }, "measuredEnergy": [{"energy": 12304, "measurement_time": "2018-03-15T10:30:00+00:00"}, {"energy": 8568, "measurement_time": "2018-03-15T12:30:00+00:00"}, {"energy": 63456, "measurement_time": "2018-03-15T14:30:00+00:00"} ] }',
            "base_url/second": '{"count": 1, "previous": "/first", "next": null, "device": {"manufacturer": "Siemens", "model": "ABC-123", "serial_number": "345345345", "latitude": "54.443567", "longitude": "-23.312543", "energy_unit": "watt_hour", "is_accumulated": false }, "measuredEnergy": [{"energy": 0, "measurement_time": "2018-03-15T16:30:00+00:00"}, {"energy": 265, "measurement_time": "2018-03-15T18:30:00+00:00"} ] }'
        }
        data = json.loads(raw[url])
        return self._parse_source(str(raw), data)


class BondAPIv1TestDevice3(BondAPIv1):
    """
    Data parsing test fixture
    """

    def _reach_source(self, url, start=None, end=None, have_next=False) -> (str, dict):
        raw = {
            "base_url/produced/0": '{"count": 0, "previous": null, "next": null, "device": {"manufacturer": "Siemens", "model": "ABC-123", "serial_number": "345345345", "latitude": "54.443567", "longitude": "-23.312543", "energy_unit": "megawatt_hour", "is_accumulated": false }, "measuredEnergy": [{"energy": 12304, "measurement_time": "2018-03-15T10:30:00+00:00"}, {"energy": 8568, "measurement_time": "2018-03-15T12:30:00+00:00"}, {"energy": 6, "measurement_time": "2018-03-15T14:30:00+00:00"} ] }'
        }
        data = json.loads(raw[url])
        return self._parse_source(str(raw), data)


if __name__ == '__main__':

    d1 = BondAPIv1TestDevice1("base_url", "produced", 0)
    print(d1.read_state().to_dict())

    d1 = BondAPIv1TestDevice2("base_url", "produced", 0)
    print(d1.read_state().to_dict())

    d1 = BondAPIv1TestDevice3("base_url", "produced", 0)
    print(d1.read_state().to_dict())
